/*
 * Project name:
     Multitasking (Simple multitasks led blinking project):
     PORTB (ouput):
     1. Timers (F0, F1)
     2. PWM (F2)
     3. Traffic Lights (F5, F6, F7)
     PORTC (input):
     4. Tasks control for traffic lights (F0, F1) (start/stop)
 * Copyright:
     (c) Stevan Tosic, 2014.
 * Description:
     This is a simple multitasking project. It turns on/off LEDs connected to
     PORTB driven by task and timer (with or without interrupt).
 * Features:
     - this is style of programming, not library
     - tasks should not block program execution by waiting on event(s)
       (state-machine controls task(s) execution)
     - processor independent (PORTABLE to any C compiler) 
       with BSP implementation
     - relative low memory consumption
     - state-machine oriented
     - cooperative
     - Board Support Package (processor) code are signed with bsp word
     - there are no scheduler or queues
     
 * Test configuration:
     MCU:             PIC16F887 or any other processor
                      http://ww1.microchip.com/downloads/en/DeviceDoc/41291F.pdf
     Dev.Board:       EasyPIC5 - ac:LEDs
                      http://www.mikroe.com/eng/products/view/297/easypic6-development-system/
     Oscillator:      HS, 08.0000 MHz
     Ext. Modules:    -
     SW:              mikroC PRO for PIC
                      http://www.mikroe.com/eng/products/view/7/mikroc-pro-for-pic/
 * NOTES:
     - Turn ON the PORT LEDs at SW6.
*/

#define NTASKS 8
#define NULL 0

typedef struct
{
   unsigned char B0:1;
   unsigned char B1:1;
   unsigned char B2:1;
   unsigned char B3:1;
   unsigned char B4:1;
   unsigned char B5:1;
   unsigned char B6:1;
   unsigned char B7:1;
} PORT;

char PORTAlias at PORTB;
PORT PORTC_OLD;

// semaphore lights - BSP bits
// b - "b"it
sbit bTask0 at PORTB.B0;
sbit bTask1 at PORTB.B1;
sbit bPWM at PORTB.B2;
sbit bRed at PORTB.B7;
sbit bYellow at PORTB.B6;
sbit bGreen at PORTB.B5;

sbit bSemStop at PORTC.B0;
sbit bSemStart at PORTC.B1;
sbit bSemStopOld at PORTC_OLD.B0;
sbit bSemStartOld at PORTC_OLD.B1;

unsigned cnt;

struct task
{
   unsigned char id;
   unsigned char  state;
   int tick;
   int interval;
} tasks[NTASKS];

// ftask "timer's" states: TICKING, TICKED
// traffic lights sequence: RED, REDYELLOW, GREEN, YELLOW
// PWM: OFF, ON,
// all (by need): START, STOP, OFF, ON, IDLE
enum { START, STOP, OFF, ON, TICKING, TICKED, RED, REDYELLOW, GREEN, YELLOW, IDLE };

void ticker()
{
     int i;
     for(i=0; i<NTASKS; i++)
     {
        if(tasks[i].tick > 0) tasks[i].tick--;
     }
}

// BSP interrupt - 2.5 ms
void interrupt() {
  if (TMR0IF_bit) {
    cnt++;                 // increment counter
    TMR0IF_bit = 0;        // clear TMR0IF
    TMR0   = 96;
  }
}

void ftask0(struct task *pt)
{
     switch(pt->state){
        case TICKING: 
            if(pt->tick > 0) break;
            else pt->state=TICKED;
        case TICKED:
            //PORTB.F0=~PORTB.F0;  // BSP
            bTask0=~bTask0;
            pt->tick=pt->interval;
            pt->state=TICKING;
            break;
        case IDLE:    break;
        default:      break;
     }
}

void ftask1(struct task *pt)
{
     switch(pt->state){
        case TICKING:
            if(pt->tick > 0) break;
            else pt->state=TICKED;
        case TICKED:
            //PORTB.F1=~PORTB.F1;  // BSP
            bTask1=~bTask1;
            pt->tick=pt->interval;
            pt->state=TICKING;
            break;
        case IDLE:    break;
        default:      break;
     }
}

void pwm(struct task *pt)
{
     switch(pt->state){
        case ON:
            if(pt->tick > 0) break;
            else {
               //PORTB.F2=1;    // BSP
               //(&PORTB)->F2=1;  // just for test
               bPWM=1;
               pt->tick=100;
               pt->state=OFF;
            }
            break;
        case OFF:
            if(pt->tick > 0) break;
            else {
               //PORTB.F2=0; // BSP
               bPWM=0;
               pt->tick=300;
               pt->state=ON;
            }
        case IDLE:    break;
        default:      break;
     }
}

void traffic_lights(struct task *pt)
{
     switch(pt->state){
        case RED:
            if(pt->tick > 0) break;
            else {
               //PORTB.F7=1; PORTB.F6=0; // BSP
               bRed=1; bYellow=0;
               pt->tick=300;
               pt->state=REDYELLOW;
            }
            break;
        case REDYELLOW:
            if(pt->tick > 0) break;
            else {
               //PORTB.F6=1; // BSP
               bYellow=1;
               pt->tick=100;
               pt->state=GREEN;
            }
        case GREEN:
            if(pt->tick > 0) break;
            else {
               //PORTB.F7=0; PORTB.F6=0; PORTB.F5=1;  // BSP
               bRed=0; bYellow=0; bGreen=1;
               pt->tick=300;
               pt->state=YELLOW;
            }
            break;
        case YELLOW:
            if(pt->tick > 0) break;
            else {
               //PORTB.F5=0; PORTB.F6=1;  // BSP
               bGreen=0; bYellow=1;
               pt->tick=100;
               pt->state=RED;
            }
            break;
        case IDLE:    break;
        default:      break;
     }
}

// start/stop semaphore by PORTC, pin1/pin0
void port_poll(struct task *pt)
{
    switch(pt->state)
    {
       case IDLE:
          if(bSemStopOld==0 && bSemStop==1) {
             bRed=bYellow=bGreen=0;
             tasks[2].tick=0;
             tasks[2].state=IDLE;
          }
          if(bSemStartOld==0 && bSemStart==1) {
             tasks[2].tick=tasks[2].interval;
             tasks[2].state=RED;
          }
          break;
       default: break;
    }
}

void task_init(int i, int state, int tick, int counter)
{
  tasks[i].id=i;
  tasks[i].state=state;
  tasks[i].tick=tick;
  tasks[i].interval=tasks[i].tick;
}

// BSP init
void bsp_init(void)
{
  OPTION_REG = 0x84;       // Assign prescaler to TMR0
  ANSEL  = 0;              // Configure AN pins as digital
  ANSELH = 0;
  C1ON_bit = 0;            // Disable comparators
  C2ON_bit = 0;
  TRISB = 0;               // PORTB is output
  PORTB = 0x00;            // Initialize PORTB
  TRISC = 0x03;            // B0 and B1 of PORTC are inputs
  TMR0  = 96;              // Timer0 initial value
  INTCON = 0xA0;           // Enable TMRO interrupt
                           // should be changed if interrupt would not be used
}

void main() {
  cnt = 0;                 // Initialize cnt
  bsp_init();              // BSP Init
  // Init task data and state, not run tasks
  task_init(0, TICKING, 100, 0);
  task_init(1, TICKING, 200, 0);
  task_init(2, RED, 0, 0);
  task_init(3, OFF, 0, 0);
  task_init(4, IDLE, 10, 0);
  
  // tasks running forever
  while(1) {
    ftask0(&tasks[0]);
    ftask1(&tasks[1]);
    traffic_lights(&tasks[2]);
    pwm(&tasks[3]);
    port_poll(&tasks[4]);
    if (cnt >= 4) {
      ticker();
      cnt = 0;                       // Reset cnt
      memcpy(&PORTC_OLD, &PORTC, 1); // BSP
    }
    // or without interrupts
    // delay_us(2500);      // 2.5 msecond delay
    // ticker();
  } // Endless loop
}